import types

class ErrorHandler(type):
   def __new__(cls, name, bases, attrs):
      for attr_name, attr_value in attrs.items():
         if isinstance(attr_value, types.FunctionType):
            attrs[attr_name] = cls._error_handler(attr_value)
      return super(ErrorHandler, cls).__new__(cls, name, bases, attrs)

   @classmethod
   def _error_handler(cls, decorated):
      def wrapper(self, *args, **kwargs):
        try:
            return decorated(self, *args, **kwargs)
        except Exception as exc:
            if self.ctx.debug:
                import traceback
                traceback.print_exc()
            self.ctx.log(str(exc))
      return wrapper



import os
import json


class Credentials(object):
    """Manage credentials on user computer."""
    def __init__(self, ctx):
        super(Credentials, self).__init__()
        self.ctx = ctx
        self.creds_path = self._path_to_credentials()
        self.creds_file = os.path.join(self.creds_path, 'auger.json')
        self.organization = None
        self.username = None
        self.api_url = None
        self.token = None

    def load(self):
        if 'AUGER_CREDENTIALS' in os.environ:
            content = os.environ.get('AUGER_CREDENTIALS', None)
            content = json.loads(content) if content else {}
        else:
            self._ensure_credentials_file()
            with open(self.creds_file, 'r') as file:
                content = json.loads(file.read())

        self.username = content.get('username')
        self.organization = content.get('organization')
        self.api_url = content.get('url', 'https://app.auger.ai')
        self.token = content.get('token')

        return self

    def save(self):
        self._ensure_credentials_file()

        content = {}
        content['username'] = self.username
        content['url'] = self.api_url
        content['token'] = self.token
        content['organization'] = self.organization

        with open(self.creds_file, 'w') as file:
            file.write(json.dumps(content))

    def verify(self):
        if self.token is None:
            raise Exception(
                'Please provide your credentials to Auger...')
        return True

    def _path_to_credentials(self):
        default_path = '%s/.augerai' % os.environ.get('HOME', os.getcwd())
        creds_path = os.path.abspath(self.ctx.config.get(
            'path_to_credentials', default_path))
        return os.environ.get('AUGER_CREDENTIALS_PATH', creds_path)

    def _ensure_credentials_file(self):
        if not os.path.exists(self.creds_path):
            os.makedirs(self.creds_path)

        if not os.path.exists(self.creds_file):
            with open(self.creds_file, 'w') as f:
                f.write('{}')



import importlib
from concurrent.futures import ThreadPoolExecutor, as_completed, thread

class ProviderRunner(object):
    """Runner executes provider jobs on threads."""

    def __init__(self, ctx, provider = None):
        super(ProviderRunner, self).__init__()
        self.ctx = ctx
        try:
            self.providers = self._load_providers(provider)
        except Exception as e:
            import traceback
            traceback.print_exc()
            raise e

    def execute(self, operation_name, *args, **kwargs):
        if len(self.providers) == 0:
            raise Exception('Please specify list of providers '
                'at config.yaml/providers')
        # if there is single operation requested
        # no need to run it on the thread
        if len(self.providers) == 1:
            p = list(self.providers.keys())[0]
            try:
                result = getattr(self.providers[p], operation_name)(*args, **kwargs)
                return {p: {
                    'result': True,
                    'data': result}}
            except Exception as e:
                return {p: {
                    'result': False,
                    'data': str(e) }}

        with ThreadPoolExecutor(max_workers=len(self.providers)) as executor:
            futures = {
                executor.submit(getattr(
                    self.providers[p], operation_name), *args, **kwargs): p
                for p in self.providers.keys()}

            try:
                results = {}
                for future in as_completed(futures):
                    try:
                        result = future.result()
                        results[futures[future]] = {
                            'result': True,
                            'data': result }
                    except Exception as e:
                        results[futures[future]] = {
                            'result': False,
                            'data': str(e) }
                return results
            except KeyboardInterrupt:
                # not a graceful termination
                executor._threads.clear()
                thread._threads_queues.clear()
                raise

    def _load_providers(self, provider = None):
        def get_instance(p):
            module = importlib.import_module('a2ml.api.%s.a2ml' % p)
            provider_class = getattr(module, '%sA2ML' % p.capitalize())
            return provider_class(self.ctx.copy(p))
        return {p: get_instance(p) for p in self.ctx.get_providers(provider)}



from a2ml.api.utils.crud_runner import CRUDRunner
from a2ml.api.utils.error_handler import ErrorHandler

class A2MLProject(metaclass=ErrorHandler):

    def __init__(self, ctx, provider):
        super(A2MLProject, self).__init__()
        self.ctx = ctx
        self.runner = CRUDRunner(ctx, provider, 'project')

    def list(self):
        return self.runner.execute('list')

    def create(self, name):
        return self.runner.execute('create', name)

    def delete(self, name):
        return self.runner.execute('delete', name)

    def select(self, name):
        return self.runner.execute('select', name)



import click
from a2ml.api.a2ml import A2ML
from a2ml.api.utils.context import pass_context


@click.command('predict', short_help='Predict with deployed model.')
@click.option('--provider', '-p', type=click.STRING, required=False,
    help='Cloud AutoML Provider.')
@click.argument('filename', required=True, type=click.STRING)
@click.option('--threshold', '-t', default=None, type=float,
    help='Threshold.')
@click.option('--model-id', '-m', type=click.STRING, required=False,
    help='Deployed model id.')
@click.option('--locally', is_flag=True, default=False,
    help='Predict locally using Docker image to run model.')
@pass_context
def cmdl(ctx, provider, filename, model_id, threshold, locally):
    """Predict with deployed model."""
    ctx.setup_logger(format='')
    A2ML(ctx, provider).predict(filename, model_id, threshold, locally)



import time

class TestTask(object):
    """TestTask"""
    def __init__(self, ctx):
        super(TestTask, self).__init__()
        self.ctx = ctx

    def iterate(self):
        self.ctx.log('Starting test task...')
        for x in range(5):
            self.ctx.log('Iteration %s' % x)
            time.sleep(2)
            self.ctx.log('Iteration %s done' % x)
